/* -----------------------------------------------------------------------------
*
* bxtptinfo: dump partition/geometry info
* 
* Note: the relevant code was taken from sfdisk, which is part of the
*       util-linux package (Copyright (C) 1995  Andries E. Brouwer (aeb@cwi.nl))
* 
* $Author: bablokb $
* $Revision: 1.12 $
*
* License: GPL2
* ----------------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>		/* atoi, free  */
#include <stdarg.h>		/* varargs  */
#include <unistd.h>		/* read, write  */
#include <fcntl.h>		/* O_RDWR  */
#include <errno.h>		/* ERANGE  */
#include <string.h>		/* index() */
#include <libgen.h>		/* basename() */
#include <ctype.h>
#include <getopt.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include <linux/unistd.h>	/* _syscall  */

#include "bxtptinfo.h"

static char gShortOpts[] = ":gbisp:qvVh";
static const struct option long_opts[] = {
    { "geometry",         no_argument, NULL, 'g' },
    { "bochsrc",          no_argument, NULL, 'b' },
    { "info",             no_argument, NULL, 'i' },
    { "size",             no_argument, NULL, 's' },
    { "partition",  required_argument, NULL, 'p' },
    { "quiet",            no_argument, NULL, 'q' },
    { "verbose",          no_argument, NULL, 'v' },
    { "version",          no_argument, NULL, 'V' },
    { "help",	          no_argument, NULL, 'h' },
    { NULL, 0, NULL, 0 }
};

int gExitStatus = 0;
int gOptQuiet = 1;
int gOptGeo = 0;
int gOptSize = 0;
int gOptBochsrc = 0;
int gOptInfo = 0;
int gPartNr = 0;

struct geometry gImageGeo;
struct sector *gSectorHead;
int    gFD;
struct disk_desc gPartInfo;  
char *gProgname;
char *gImageName;
off_t gImageSize;

static void error(char *s, ...) {
    va_list p;
    va_start(p, s);
    fflush(stdout);
    fprintf(stderr, "\n%s: ",gProgname);
    vfprintf(stderr, s, p);
    va_end(p);
}

static void fatal(char *s, ...) {
    va_list p;
    va_start(p, s);
    fflush(stdout);
    fprintf(stderr, "\n%s: ",gProgname);
    vfprintf(stderr, s, p);
    va_end(p);
    exit(1);
}

/*
 * sseek: seek to specified sector - return 0 on failure
 *
 * For >4GB disks lseek needs a > 32bit arg, and we have to use llseek.
 * On the other hand, a 32 bit sector number is OK until 2TB.
 * The routines _llseek and sseek below are the only ones that
 * know about the loff_t type.
 *
 * Note: we use 512-byte sectors here, irrespective of the hardware ss.
 */
#if !defined (__alpha__) && !defined (__ia64__) && !defined (__x86_64__)
static
_syscall5(int,  _llseek,  uint,  fd, ulong, hi, ulong, lo,
       loff_t *, res, uint, wh);
#endif

static int
sseek(char *dev, unsigned int fd, unsigned long s) {
    loff_t in, out;
    in = ((loff_t) s << 9);
    out = 1;

#if !defined (__alpha__) && !defined (__ia64__) && !defined (__x86_64__)
    if (_llseek (fd, in>>32, in & 0xffffffff, &out, SEEK_SET) != 0) {
#else
    if ((out = lseek(fd, in, SEEK_SET)) != in) {
#endif
	perror("llseek");
	error("seek error on %s - cannot seek to %lu\n", dev, s);
	return 0;
    }

    if (in != out) {
	error("seek error: wanted 0x%08x%08x, got 0x%08x%08x\n",
	       (uint)(in>>32), (uint)(in & 0xffffffff),
	       (uint)(out>>32), (uint)(out & 0xffffffff));
	return 0;
    }
    return 1;
}


static void version(void) {
    printf("\n\n%s: query partition information, version %s\n\n"
	   " Copyright (C) 2003 Bernhard Bablok, mail@bablokb.de\n"
           " This is free software; see the source for copying conditions.\n"
           " There is NO warranty; not even for MERCHANTABILITY or\n"
           " FITNESS FOR A PARTICULAR PURPOSE.\n\n", gProgname, VERSION);
}

static void usage(void) {
    version();
    printf("Usage: %s [options] image\n", gProgname);
    puts ("image: image created with bximage or bxtcreate");
    puts ("options:");
    puts ("    -g [or --geometry]:     output geometry");
    puts ("    -b [or --bochsrc]:      output geometry (bochsrc-format)");
    puts ("    -i [or --info]:         output image info in human-readable form");
    puts ("    -s [or --size]:         output size");
    puts ("    -p# [or --partition #]: output partition info");
    puts ("    -q [or --quiet]:        suppress warnings");
    puts ("    -V [or --version]:      print version");
    puts ("    -h [or --help]:         print this message");
    exit(1);
}


static void free_sectors(void) {
    struct sector *s;

    while (gSectorHead) {
	s = gSectorHead;
	gSectorHead = s->next;
	free(s);
    }
}

static struct sector * get_sector(char *dev, int fd, unsigned long sno) {
    struct sector *s;

    for(s = gSectorHead; s; s = s->next)
	if(s->sectornumber == sno)
	    return s;

    if (!sseek(dev, fd, sno))
	return 0;

    if (!(s = (struct sector *) malloc(sizeof(struct sector))))
	fatal("out of memory - giving up\n");

    if (read(fd, s->data, sizeof(s->data)) != sizeof(s->data)) {
	if (errno)		/* 0 in case we read past end-of-disk */
	    perror("read");
	error("read error on %s - cannot read sector %lu\n", dev, sno);
	free(s);
	return 0;
    }

    s->next = gSectorHead;
    gSectorHead = s;
    s->sectornumber = sno;
    s->to_be_written = 0;

    return s;
}

static longchs chs_to_longchs (chs a) {
    longchs aa;

    aa.h = a.h;
    aa.s = (a.s & 0x3f);
    aa.c = (a.s & 0xc0);
    aa.c = (aa.c << 2) + a.c;
    return aa;
}

static void setGeometry(long cyl, long heads, long spt) {
  gImageGeo.heads = heads;
  gImageGeo.sectors = spt;
  gImageGeo.cylindersize = heads*spt;
  if (cyl == 0 && gImageGeo.cylindersize != 0)
    gImageGeo.cylinders = gImageSize/512/gImageGeo.cylindersize;
  else
    gImageGeo.cylinders = cyl;
}

static struct geometry get_fdisk_geometry_one(struct part_desc *p) {
    struct geometry G;

    chs b = p->p.end_chs;
    longchs bb = chs_to_longchs(b);
    G.heads = bb.h+1;
    G.sectors = bb.s;
    G.cylindersize = G.heads*G.sectors;
    G.cylinders = G.start = 0;
    return G;
}

static void get_fdisk_geometry(struct disk_desc *z) {
    struct part_desc *p;
    int pno, agree;
    struct geometry G0, G;

    agree = 0;
    G0.heads = G0.sectors = 0;
    for (pno=0; pno < z->partno; pno++) {
	p = &(z->partitions[pno]);
	if (p->size != 0 && p->p.sys_type != 0) {
	    G = get_fdisk_geometry_one(p);
	    if (!G0.heads) {
		G0 = G;
		agree = 1;
	    } else if (G.heads != G0.heads || G.sectors != G0.sectors) {
		agree = 0;
		break;
	    }
	}
    }
    setGeometry(G0.cylinders,G0.heads,G0.sectors);
}

static int copy_to_int(unsigned char *cp) {
    unsigned int m;

    m = *cp++;
    m += (*cp++ << 8);
    m += (*cp++ << 16);
    m += (*cp++ << 24);
    return m;
}

static void copy_to_part(char *cp, struct partition *p) {
    p->bootable = *cp++;
    p->begin_chs.h = *cp++;
    p->begin_chs.s = *cp++;
    p->begin_chs.c = *cp++;
    p->sys_type = *cp++;
    p->end_chs.h = *cp++;
    p->end_chs.s = *cp++;
    p->end_chs.c = *cp++;
    p->start_sect = copy_to_int(cp);
    p->nr_sects = copy_to_int(cp+4);
}

static int msdos_signature (struct sector *s) {
  unsigned char *data = s->data;
  if (data[510] == 0x55 && data[511] == 0xaa)
    return 1;
  else
    return 0;
}


static int msdos_partition(char *dev, int fd, unsigned long start,
                                                    struct disk_desc *z) {
    char *cp;
    struct partition pt;
    struct sector *s;
    struct part_desc *partitions = &(z->partitions[0]);
    int pno = z->partno;

    if (!(s = get_sector(dev, fd, start)))
	return 0;

    if (!msdos_signature(s))
	return 0;

    cp = s->data + 0x1be;
    copy_to_part(cp,&pt);

    /* If I am not mistaken, recent kernels will hide this from us,
	   so we will never actually see traces of a Disk Manager */
    if (pt.sys_type == DM6_PARTITION
	|| pt.sys_type == EZD_PARTITION
	|| pt.sys_type == DM6_AUX1PARTITION
	|| pt.sys_type == DM6_AUX3PARTITION) {
	fatal("detected Disk Manager - unable to handle that\n");
	return 0;
    }
    { unsigned int sig = *(unsigned short *)(s->data + 2);
      if (sig <= 0x1ae
	  && *(unsigned short *)(s->data + sig) == 0x55aa
	  && (1 & *(unsigned char *)(s->data + sig + 2))) {
	  fatal("DM6 signature found - giving up\n");
	  return 0;
      }
    }

    for (pno=0; pno<4; pno++,cp += sizeof(struct partition)) {
	partitions[pno].sector = start;
	partitions[pno].offset = cp - s->data;
	copy_to_part(cp,&pt);
	partitions[pno].start = start + pt.start_sect;
	partitions[pno].size = pt.nr_sects;
	partitions[pno].ep = 0;
	partitions[pno].p = pt;
    }

    z->partno = pno;
    return 1;
}

static void get_partitions(char *dev, int fd, struct disk_desc *z) {
    z->partno = 0;
    msdos_partition(dev, fd, 0, z);
}

static int my_open (char *dev, int rw) {
    int fd, mode;

    mode = (rw ? O_RDWR : O_RDONLY);
    fd = open(dev, mode);
    if (fd < 0) {
      perror(dev);
      if (rw)
	fatal("cannot open %s read-write\n", dev);
      else
	fatal("cannot open %s for reading\n", dev);
    }
    return fd;
}

static void getImageSize() {
  struct stat image_stats;
  if (stat(gImageName, &image_stats)) {
    fatal("%s: %s\n",gImageName,strerror(errno));
  }
  gImageSize = image_stats.st_size;
}

static int tryGeoPartition() {
  if (gPartInfo.partno == 0)
    return 0;
  else
    get_fdisk_geometry(&gPartInfo);
  return gImageGeo.cylinders != 0;
}

static int tryGeoWellKnown() {

  /* some known images-sizes: see "old" bochs-documentation */

  switch(gImageSize) {
      case 10653696: setGeometry(306,4,17); return 1;
      case 21411840: setGeometry(615,4,17); return 1;
      case 32117760: setGeometry(615,6,17); return 1;
      case 49090560: setGeometry(940,6,17); return 1;
      case 65454080: setGeometry(940,8,17); return 1;
     case 117504000: setGeometry(900,15,17); return 1;
     case 495452160: setGeometry(1024,15,63); return 1;
  default:
    break;
  }

  /* bximage/bxtcreate compatible geometry: 16 heads, 63 spt */

  if ((gImageSize/512)%(16*63) == 0) {
    setGeometry(gImageSize/512/16/63,16,63);
    return 1;
  }
  return 0;
}

static int tryGeoGuess() {
  long spt, heads, size = gImageSize/512, size2;
  for (spt = 63; spt > 1; --spt) {
    if (size%spt != 0)
      continue;
    size2 = size/spt;
    heads = 2;
    while (heads*heads <= size2) {
      if (size2%heads == 0) {
	setGeometry(size2/heads,heads,spt);
	return 1;
      }
      heads++;
    }
  }
  return 0;
}

static void read_partitions() {
  gFD = my_open(gImageName,0);
  free_sectors();
  get_partitions(gImageName,gFD,&gPartInfo);
  setGeometry(0,0,0);
  tryGeoPartition() || tryGeoWellKnown() || tryGeoGuess();
  if (gImageGeo.cylinders == 0)
    fatal("failed to detect geometry\n");
}

static void out_geometry() {
  if (gOptBochsrc)
    printf("cylinders=%lu, heads=%lu, spt=%lu\n",
	   gImageGeo.cylinders,gImageGeo.heads,gImageGeo.sectors);
  else
    printf("%lu %lu %lu\n", gImageGeo.cylinders,gImageGeo.heads,gImageGeo.sectors);
}

static void out_size(void) {
  printf("%lu\n", (long unsigned) gImageSize);
}

static void out_partInfo() {
  struct part_desc part = gPartInfo.partitions[gPartNr-1];
  printf("%u %lu %x\n",512*part.p.start_sect,part.size/2,part.p.sys_type);
}

static void out_all() {
  int i;
  unsigned long size;
  struct part_desc part;
  printf("\nHD-image:   %s\n"
         "Total size: %lu MB (%lu KB, %lu bytes)\n"
         "Geometry:   C/H/S = %lu/%lu/%lu\n\n"
         "Partitions:\n\n",
	 gImageName, (long unsigned) gImageSize/1024/1024,
	 (long unsigned) gImageSize/1024, (long unsigned) gImageSize,
	 gImageGeo.cylinders,gImageGeo.heads,gImageGeo.sectors);

  for (i=0; i < gPartInfo.partno; ++i) {
    part = gPartInfo.partitions[i];
    size = part.size/2;
    if (size == 0)
      continue;
    printf("%s:%d - size: %lu MB (%lu KB, %lu bytes), id: %x\n",
           gImageName, i+1, size/1024, size, size*1024, part.p.sys_type); 
  }
}

int main(int argc, char **argv) {
  int  c;

  gProgname = basename(strdup(argv[0]));
  if (argc < 2)
    usage();

  while ((c = getopt_long (argc, argv, gShortOpts, long_opts, NULL)) != -1) {
    switch (c) {
    case 'g':
      gOptGeo = 1; break;
    case 'b':
      gOptGeo = 1; gOptBochsrc = 1; break;
    case 'i':
      gOptInfo = 1; break;
    case 's':
      gOptSize = 1; break;
    case 'p':
      gPartNr = atoi(optarg); break;
    case 'q':
      gOptQuiet = 1; break;
    case 'v':
      gOptQuiet = 0; break;
    case 'V':
      version();
      exit(0);
    case '?':
    default:
      error("illegal option: -%c",optopt);
      usage(); break;
    }
  }

  gImageName = argv[optind];
  getImageSize();
  read_partitions();

  if (gOptInfo) {
    out_all();
    return 0;
  }

  if (gOptSize)
    out_size();
  if (gOptGeo)
    out_geometry();
  if (gPartNr)
    out_partInfo();
  exit(gExitStatus);
  return 0;
}
